import typing
if typing.TYPE_CHECKING:
    from .core import Usys

import sqlalchemy as sa
from sqlalchemy import Engine, create_engine

from pyuu.common import *
from pyuu.trivial import *
from pyuu.path import Path
from pyuu.env import AppEnv, SysEnv
from pyuu.project import ProjPath


class PkgMngrBase:
    def __init__(self, usys:'Usys', PATH:ProjPath):
        self.usys=usys
        self.PATH=PATH

    def register(self, update):
        pass





import typing
if typing.TYPE_CHECKING:
    from ..core import Usys
from .common import *

__all__=['PkgService']

class PkgService:
    def __init__(self, usys:'Usys'):
        self.ctx = usys.DbCtx

    @property
    def TABLE(self):
        return self.ctx.core.TABLE
    
    @property
    def engine(self):
        return self.ctx.core.engine

    def get_pkg_path(self, pkg_id, version):
        pass

    def begin(self):
        return self.engine.begin()

    def register_pkg(self, pkg_id, version, path, *, update=True, conn:sa.Connection=None):
        TABLE=self.TABLE
        engine=self.engine

        version=nvl(version, '')

        def impl(conn:sa.Connection):
            stmt=sa.insert(TABLE.Pkg).values(pkg_id=pkg_id, version=version, path=str(path))
            if update:
                self.unregister_pkg(pkg_id, version, conn=conn)
                conn.execute(stmt)
            else:
                conn.execute(stmt)

        if conn is None:
            with self.begin() as conn:
                impl(conn)
        else:
            impl(conn)


    def unregister_pkg(self, pkg_id, version='', *, conn:sa.Connection=None):
        TABLE=self.TABLE
        engine=self.engine
        Pkg=TABLE.Pkg

        version=nvl(version, '')

        def impl(conn:sa.Connection):
            stmt=sa.delete(Pkg).where(util.make_stmt_eqs(TABLE.get_keys_map(TABLE.Pkg), dict(pkg_id=pkg_id, version=version)))
            conn.execute(stmt)

        if conn is None:
            with self.begin() as conn:
                impl(conn)
        else:
            impl(conn)
        





from .common import *
from sqlalchemy import Column, Integer, MetaData, Table, Text, and_


def create_sqlite_engine(pth:Path):
    if logger.level <=logging.DEBUG:
        debug=True

    pth.prnt.mkdir()
    engine = create_engine(f'sqlite:///{pth}',echo=debug)
    with engine.begin() as conn:
        conn.exec_driver_sql('PRAGMA journal_mode = WAL;')
        conn.exec_driver_sql('PRAGMA synchronous = NORMAL;')
    return engine


def make_cols_map(c, names):
    return {name:c.get(name) for name in names}

def make_core_tables(metadata:MetaData=None):
    if metadata is None:
        metadata = MetaData()

    class TABLE:
        Pkg = Table(
            'Pkg', metadata,
            Column('rid', Integer, primary_key=True, autoincrement=True),
            Column('pkg_id', Text, nullable=False, index=True),
            Column('version', Text, nullable=False,),
            Column('path', Text)
        )

        __keys={
            id(Pkg):make_cols_map(Pkg.c, ['pkg_id', 'version'])
        }

        @staticmethod
        def get_keys_map(pkg):
            return TABLE.__keys[id(pkg)]

    return TABLE


def make_stmt_eqs(cols_map:dict, vals_map:dict):
    eqs=[]
    for k in cols_map:
        if k in vals_map:
            eqs.append(cols_map[k]==vals_map[k])
    return and_(*eqs)




import typing
import importlib
if typing.TYPE_CHECKING:
    from ..core import Usys

import sys

from pyuu.uimport import DictMetaPathFinder

from .common import *

__all__=['PkgMngrService']


class PkgModule:
    class PkgMngr(PkgMngrBase):
        pass

class UsysException(Exception):
    pass

class PkgMngrExistsException(UsysException):
    pass


class PkgMngrService:
    def __init__(self, usys:'Usys'):
        self.usys = usys
        self.ctx = usys.DbCtx
        self.pm_module_map={}
        self.pm_meta_path_finder=DictMetaPathFinder(self.pm_module_map)
        sys.meta_path.append(self.pm_meta_path_finder)

    def create_pkg_mngr_from_repo(self, pkg_id:str=None)->PkgModule.PkgMngr:
        return self.create_pkg_mngr(self.get_pkg_mngr_mod_path(pkg_id), pkg_id)

    def create_pkg_mngr(self, pkg_path:Path, pkg_id:str=None)->PkgModule.PkgMngr:
        assert pkg_path.exists()
        pkg_id = nvl(pkg_id, pkg_path.stem)
        full_module_name=f'pyuu.usys.{pkg_id}'
        if pkg_path_exists := self.pm_module_map.get(full_module_name):
            if pkg_path_exists is not None and Path(pkg_path_exists).resolve()!=pkg_path.resolve():
                raise PkgMngrExistsException
        self.pm_module_map[full_module_name]=pkg_path
        m:PkgModule = importlib.import_module(full_module_name)
        pkg_mngr = m.PkgMngr(self.usys, self.get_mod_path(pkg_id))
        return pkg_mngr
    
    def get_pkg_mngr_mod_path(self, pkg_id):
        pkg_mngr_path=self.usys.PATH.pkg_mngr
        p:Path=pkg_mngr_path/f'{pkg_id}.py'
        if p.exists():
            return p
        p=pkg_mngr_path/pkg_id
        if p.exists():
            return p
    
    def save_pkg_mngr_mod(self, pkg_path:Path):
        pkg_mngr_path=self.usys.PATH.pkg_mngr
        pkg_path.copy_to(pkg_mngr_path)
        
    def get_mod_path(self, name):
        P=self.usys.PATH
        class PATH:
            assets:Path=P.pkg_mngr/name
            data:Path=P.data/'mod/pkg_mngr/data'/name
            cache:Path=P.cache/'mod/pkg_mngr'/name
            tmp:Path=P.tmp/'mod/pkg_mngr'/name
            out:Path=P.out/'mod/pkg_mngr'/name
            export:Path=P.pkg_mngr/name/'export'
        return PATH

    def __del__(self):
        if sys.meta_path:
            sys.meta_path.remove(self.pm_meta_path_finder)





from sqlalchemy import Engine, MetaData
from .common import *
from .util import *
from .service import *

__all__=['Usys']



class Usys(AppEnv):
    name='usys'
    def __init__(self, root_path:Path=None):
        self.sys_env = sys_env = SysEnv()
        root_path = nvl(root_path, sys_env.vars['USYS_PATH'])
        assert root_path

        name = self.name
        root_path = Path(root_path)
        class PATH:
            root = root_path
            assets = root/'assets'
            data = root/'data'
            cache = root/'cache'
            tmp = root/'tmp'
            out = root/'out'

            messey = tmp/'messey'
            export = root/'export'
            test = root/'test'
            test_data = test/'data'

            uu = root/'.uu'             # .uu to indicate root

            secret=data/'secret'
            mod=data/'mod'              # module
            pkg_mngr=mod/'pkg_mngr'/'assets'    
            pm=mod/'pm'                 # app manager

            # ALL_DIRS = [root, assets, data, cache, tmp, out, export, uu, pm]
        self.PATH=PATH

        class DbCtx:
            class core:
                engine=create_sqlite_engine(PATH.data/'core/db.sqlite')
                metadata=MetaData()
                TABLE=make_core_tables(metadata)
                metadata.create_all(engine)

        self.DbCtx=DbCtx
        self.pkg_service=PkgService(self)
        self.pkg_mngr_service=PkgMngrService(self)

        


